/*
 * Thumb-1 functions.
 *
 * Copyright 2016 Ling LI <lix2ng@gmail.com>, under the MIT license.
 *
 * Cycle counts are referencing Cortex-M0+ with 1-cycle multiplier. Caller's
 * cost is not included (that's 4 bytes and 3 cycles for each BL).
 *
 * arm-none-eabi-gcc -mthumb -mcpu=cortex-m0 -masm-syntax-unified -Os -c thumb1.c
 * Add -fshort-wchar if necessary (use with Keil for example).
 */

#include <stdint.h>
#include "thumb1.h"

#if defined(__thumb__) && !defined(__thumb2__)

#define _naked __attribute__((naked))
#define _sect(x) __attribute__((section(x)))
#define br "\n\t"

/*
 * Size: 52B; cycles: 16~19.
 * Uses r0, r1, r2.
 */
_naked _sect(".text.thumb1_clz")
int thumb1_clz(uint32_t x)
{
    __asm(
        ".global thumb1_clznz" br
        "thumb1_clznz:" br
        "movs   r1, #28" br
        "lsrs   r2, r0, #16" br
        "beq    1f" br
        "subs   r1, #16" br
        "mov    r0, r2" br
        "1:" br
        "lsrs   r2, r0, #8" br
        "beq    2f" br
        "subs   r1, #8" br
        "mov    r0, r2" br
        "2:" br
        "lsrs   r2, r0, #4" br
        "beq    3f" br
        "subs   r1, #4" br
        "mov    r0, r2" br
        "3:" br
        "adr    r2, .L_clzlut" br
        "ldrb   r0, [r2, r0]" br
        "adds   r0, r1" br
        "bx     lr" br
        ".align 2" br
        ".L_clzlut:" br
        ".byte 4, 3, 2, 2, 1, 1, 1, 1" br
        ".byte 0, 0, 0, 0, 0, 0, 0, 0"
    );
}

/*
 * Size: 60B; cycles: 13 (6 when input is 0).
 * Alternative entry for non-zero input: 11 cycles.
 * Uses r0, r1.
 */
_naked _sect(".text.thumb1_ctz")
int thumb1_ctz(uint32_t x)
{
    __asm(
        "cmp    r0, #0" br
        "beq    1f" br

        ".thumb_func" br
        ".global thumb1_ctznz" br

        "thumb1_ctznz:" br
        "rsbs   r1, r0, #0" br
        "ands   r0, r1" br      // isolate the lsb1.
        "ldr    r1, =0x077cb531" br // de Bruijn sequence.
        "muls   r0, r1" br
        "adr    r1, .L_ctzlut" br
        "lsrs   r0, #27" br     // top 5 bits as index.
        "ldrb   r0, [r1, r0]" br
        "bx     lr" br
        "1:" br
        "movs   r0, #32" br
        "bx     lr" br
        ".ltorg" br
        ".L_ctzlut:" br
        ".byte 0, 1, 28, 2, 29, 14, 24, 3" br
        ".byte 30, 22, 20, 15, 25, 17, 4, 8" br
        ".byte 31, 27, 13, 23, 21, 19, 16, 7" br
        ".byte 26, 12, 18, 6, 11, 5, 10, 9" br
    );
}

/*
 * Size: 52B; cycles: 24.
 * Uses r0, r1, r2.
 */
_naked _sect(".text.thumb1_rbit")
uint32_t thumb1_rbit(uint32_t x)
{
    __asm(
        /* Swap odd and even. */
        "ldr    r2, =0x55555555" br
        "lsrs   r1, r0, #1" br
        "lsls   r0, #1" br
        "ands   r1, r2" br
        "bics   r0, r2" br
        "orrs   r0, r1" br
        /* Swap adjacent pairs. */
        "ldr    r1, =0x33333333" br
        "lsrs   r2, r0, #2" br
        "lsls   r0, #2" br
        "ands   r2, r1" br
        "bics   r0, r1" br
        "orrs   r2, r0" br
        /* Swap adjacent nibbles. */
        "ldr    r0, =0x0f0f0f0f" br
        "lsrs   r1, r2, #4" br
        "lsls   r2, #4" br
        "ands   r1, r0" br
        "bics   r2, r0" br
        "orrs   r1, r2" br
        /* Reverse all bytes. */
        "rev    r0, r1" br
        "bx     lr" br
        ".ltorg"
    );
}

/*
 * Size: 40B; cycles: 21.
 * Uses r0, r1, r2, r3, r12.
 */
_naked _sect(".text.thumb1_umulh")
uint32_t thumb1_umulh(uint32_t a, uint32_t b)
{
    __asm(
        "uxth   r3, r1" br
        "mov    r12, r4" br
        "mov    r2, r3" br
        "uxth   r4, r0" br
        "muls   r2, r4" br
        "lsrs   r0, #16" br
        "muls   r3, r0" br
        "lsrs   r1, #16" br
        "muls   r0, r1" br
        "lsrs   r2, #16" br
        "muls   r1, r4" br
        "adds   r2, r3" br
        "mov    r4, r12" br
        "uxth   r3, r2" br
        "adds   r1, r3" br
        "lsrs   r2, #16" br
        "lsrs   r1, #16" br
        "adds   r2, r0" br
        "adds   r0, r1, r2" br
        "bx     lr"
    );
}

/*
 * Size: 40B, cycles: 21.
 * Uses r0, r1, r2, r3, r12.
 */
_naked _sect(".text.thumb1_smulh")
int32_t thumb1_smulh(int32_t a, int32_t b)
{
    __asm(
        "uxth   r2, r1" br
        "mov    r12, r4" br
        "mov    r3, r2" br
        "uxth   r4, r0" br
        "muls   r3, r4" br
        "asrs   r0, #16" br
        "muls   r2, r0" br
        "asrs   r1, #16" br
        "muls   r0, r1" br
        "asrs   r3, #16" br
        "muls   r1, r4" br
        "adds   r3, r2" br
        "mov    r4, r12" br
        "uxth   r2, r3" br
        "adds   r1, r2" br
        "asrs   r3, #16" br
        "asrs   r1, #16" br
        "adds   r3, r0" br
        "adds   r0, r1, r3" br
        "bx     lr"
    );
}

/*
 * Size: 54B total. Cycles:
 *  - udiv16: 122~138
 *  - sdiv16: 130~148
 * Uses r0, r1, r2, r3, r12.
 */
_naked _sect(".text.thumb1_div16")
int32_t thumb1_sdiv16(int32_t a, int32_t b)
{
    __asm(
        "movs   r2, #1" br
        "cmp    r0, #0" br
        "bpl    1f" br
        "rsbs   r0, #0" br      // make dividend positive.
        "rsbs   r2, #0" br
        "1:" br
        "cmp    r1, #0" br
        "bpl    2f" br
        "rsbs   r1, #0" br      // make divisor positive.
        "rsbs   r2, #0" br
        "2:" br                 // r2 is now +/-1 depending on signs of r0/r1.
        "b      .L_do_div" br

        ".thumb_func" br
        ".global thumb1_udiv16" br
        "thumb1_udiv16:" br
        "movs   r2, #1" br      // unsigned, sign set to +1.

        ".L_do_div:" br
        "mov    r12, r2" br
        "movs   r3, #1" br
        "lsls   r3, #15" br     // bit slide <- 0x8000.
        "lsls   r1, #15" br
        "movs   r2, #0" br
        "1:" br
        "cmp    r0, r1" br      // simple shift-sub loop.
        "blo    2f" br
        "subs   r0, r1" br
        "orrs   r2, r3" br
        "2:" br
        "lsrs   r1, #1" br
        "lsrs   r3, #1" br
        "bne    1b" br          // 16 rounds.
        "mov    r0, r2" br
        "mov    r3, r12" br
        "muls   r0, r3" br      // apply sign.
        "bx     lr" br
    );
}

#else
#warning "For Cortex-M0/M0+ only."

#endif

/* vim: set syn=c.doxygen cin et sw=4 ts=4 tw=80 fo=croqm: */
